#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import time
from collections import Iterable
from typing import Union, Tuple

from notification_service.model.event import ANY_CONDITION, Event
from notification_service.model.sender_event_count import SenderEventCount
from notification_service.storage.event_storage import BaseEventStorage


class MemoryEventStorage(BaseEventStorage):

    def __init__(self):
        self.store = []
        self.store_with_uuid = {}
        self.max_offset = 0
        self.clients = []

    def add_event(self, event: Event, uuid: str):
        self.max_offset += 1
        event.create_time = int(time.time() * 1000)
        event.offset = self.max_offset
        self.store.append(event)
        self.store_with_uuid[uuid] = event
        return event

    def list_events(self,
                    key: str = None,
                    namespace: str = None,
                    sender: str = None,
                    begin_offset: int = None,
                    end_offset: int = None):
        res = []
        key = None if key == "" else key
        namespace = None if namespace == "" else namespace
        sender = None if sender == "" else sender
        begin_offset = None if begin_offset == 0 else begin_offset
        end_offset = None if end_offset == 0 else end_offset

        for event in self.store:
            if key is not None and key != ANY_CONDITION and event.key != key:
                continue
            if namespace is not None and namespace != ANY_CONDITION and event.namespace != namespace:
                continue
            if sender is not None and sender != ANY_CONDITION and event.sender != sender:
                continue
            if begin_offset is not None and event.offset <= begin_offset:
                continue
            if end_offset is not None and event.offset > end_offset:
                continue
            res.append(event)
        return res

    def count_events(self,
                     key: str = None,
                     namespace: str = None,
                     sender: str = None,
                     begin_offset: int = None,
                     end_offset: int = None):
        key = None if key == "" else key
        namespace = None if namespace == "" else namespace
        sender = None if sender == "" else sender
        begin_offset = None if begin_offset == 0 else begin_offset
        end_offset = None if end_offset == 0 else end_offset
        event_counts = {}
        for event in self.store:
            if key is not None and key != ANY_CONDITION and event.key != key:
                continue
            if namespace is not None and namespace != ANY_CONDITION and event.namespace != namespace:
                continue
            if sender is not None and sender != ANY_CONDITION and event.sender != sender:
                continue
            if begin_offset is not None and event.offset <= begin_offset:
                continue
            if end_offset is not None and event.offset > end_offset:
                continue
            if event.sender in event_counts:
                event_counts.update({event.sender: event_counts.get(event.sender) + 1})
            else:
                event_counts.update({event.sender: 1})
        res = []
        for sender, event_count in event_counts.items():
            res.append(SenderEventCount(sender=sender, event_count=event_count))
        return res

    def list_all_events(self, start_time: int):
        res = []
        for event in self.store:
            if event.create_time >= start_time:
                res.append(event)
        return res

    def list_all_events_from_offset(self, start_offset: int, end_offset: int = None):
        res = []
        for event in self.store:
            if 0 < end_offset < event.offset:
                continue
            if event.offset > start_offset:
                res.append(event)
        return res

    def register_client(self, namespace: str = None, sender: str = None) -> int:
        self.clients.append((len(self.clients), namespace, sender))
        return len(self.clients) - 1

    def delete_client(self, client_id):
        try:
            self.clients.pop(client_id)
        except:
            pass

    def is_client_exists(self, client_id) -> bool:
        for item in self.clients:
            if item[0] == client_id:
                return True
        return False

    def get_event_by_uuid(self, uuid: str):
        return self.store_with_uuid.get(uuid)

    def timestamp_to_event_offset(self, timestamp: int) -> int:
        for e in self.store:
            if e.create_time > timestamp:
                return e.offset - 1
        return 0

    def clean_up(self):
        self.store.clear()
        self.clients.clear()
        self.store_with_uuid.clear()
        self.max_offset = 0



#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from abc import ABC, abstractmethod

from notification_service.model.event import Event


class BaseEventStorage(ABC):

    @abstractmethod
    def add_event(self, event: Event, uuid: str):
        pass

    @abstractmethod
    def list_events(self,
                    key: str = None,
                    namespace: str = None,
                    sender: str = None,
                    begin_offset: int = None,
                    end_offset: int = None):
        pass

    @abstractmethod
    def count_events(self,
                     key: str = None,
                     namespace: str = None,
                     sender: str = None,
                     begin_offset: int = None,
                     end_offset: int = None):
        pass

    @abstractmethod
    def clean_up(self):
        pass

    @abstractmethod
    def register_client(self, namespace: str = None, sender: str = None) -> int:
        pass

    @abstractmethod
    def delete_client(self, client_id):
        pass

    @abstractmethod
    def is_client_exists(self, client_id) -> bool:
        pass

    def get_event_by_uuid(self, uuid: str):
        pass

    def timestamp_to_event_offset(self, timestamp: int) -> int:
        pass



# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import abc
from datetime import datetime
from typing import List, Optional

from notification_service.model.event import Event


class ListenerRegistrationId(object):
    def __init__(self, id: str):
        self.id = id


class ListenerProcessor(object):
    def process(self, events: List[Event]):
        pass


class NotificationClient(metaclass=abc.ABCMeta):
    def __init__(self,
                 namespace: Optional[str] = None,
                 sender: Optional[str] = None):
        self.namespace = namespace
        self.sender = sender

    @abc.abstractmethod
    def send_event(self, event: Event):
        """
        Send event to Notification Sever.

        :param event: the event updated.
        :return: The created event.
        """
        pass

    @abc.abstractmethod
    def register_listener(self,
                          listener_processor: ListenerProcessor,
                          event_keys: List[str] = None,
                          offset: int = None
                          ) -> ListenerRegistrationId:
        """
        Register a listener to listen events from Notification Server

        :param listener_processor: The processor of the listener.
        :param event_keys: Keys of event for listening. If not set, it will listen all events.
        :param offset: The offset of the events to start listening.
        :return: The `ListenerRegistrationId` used to stop the listening.
        """
        pass

    @abc.abstractmethod
    def unregister_listener(self,
                            registration_id: ListenerRegistrationId):
        """
        Unregister the listener by id.
        """
        pass

    @abc.abstractmethod
    def list_events(self,
                    key: str = None,
                    namespace: str = None,
                    sender: str = None,
                    begin_offset: int = None,
                    end_offset: int = None) -> List[Event]:
        """
        List specific events in Notification Service.

        :param key: The key of the event to list.
        :param namespace: The namespace of the event to list.
        :param sender: The sender of the event to list.
        :param begin_offset: Offset of the events must be greater than this offset.
        :param end_offset: Offset of the events must be less than or equal to this offset.
        :return: The event list.
        """
        pass

    @abc.abstractmethod
    def time_to_offset(self, time: datetime) -> int:
        """
        Look up the biggest offset of the event whose create_time is
        less than or equal to the given timestamp.

        :return: The offset of the founded event.
        """
        pass



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Main executable module"""

import argcomplete

from notification_service.cli import cli_parser


def main():
    """Main executable function"""

    parser = cli_parser.get_parser()
    argcomplete.autocomplete(parser)
    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import logging
import os
import threading
import time
import grpc
from functools import wraps
from random import shuffle
from datetime import datetime
from typing import List, Dict, Tuple


from notification_service.model.event import Event
from notification_service.client.notification_client import NotificationClient, ListenerRegistrationId, \
    ListenerProcessor
from notification_service.rpc.protobuf import notification_service_pb2_grpc
from notification_service.rpc.protobuf.notification_service_pb2 import SendEventRequest, \
    ListEventsRequest, CountEventsRequest, EventProto, ReturnStatus, ListAllEventsRequest, \
    RegisterClientRequest, ClientMeta, ClientIdRequest, TimeToOffsetRequest, ListMembersRequest
from notification_service.util.utils import event_proto_to_event, proto_to_member, sleep_and_detecting_running, \
    event_count_proto_to_event_count

from notification_service.model.sender_event_count import SenderEventCount

NOTIFICATION_TIMEOUT_SECONDS = os.environ.get("NOTIFICATION_TIMEOUT_SECONDS", 5)


class SequenceNumberManager(object):

    def __init__(self, seq_num: int = 0):
        self._lock = threading.RLock()
        self._seq_num = seq_num

    def increment_sequence_number(self):
        with self._lock:
            self._seq_num += 1

    def get_sequence_number(self):
        return self._seq_num


class EmbeddedNotificationClient(NotificationClient):
    def __init__(self,
                 server_uri: str,
                 namespace: str = None,
                 sender: str = None,
                 client_id: int = None,
                 initial_seq_num: int = None
                 ):
        super().__init__(namespace, sender)
        self.server_uri = server_uri
        channel = grpc.insecure_channel(target=self.server_uri,
                                        options=[('grpc.max_receive_message_length', -1)])
        self.notification_stub = notification_service_pb2_grpc.NotificationServiceStub(channel)
        self.threads = {}  # type: Dict[int, threading.Thread]
        self.lock = threading.Lock()
        self._client_id = client_id
        self._initial_seq_num = initial_seq_num

        self.list_member_interval_ms = 1000
        self.retry_interval_ms = 1000
        self.retry_timeout_ms = 10000

        server_uri_list = self.server_uri.split(",")
        if len(server_uri_list) > 1:
            self.living_members = []
            self.current_uri = None
            last_error = None
            for server_uri in server_uri_list:
                channel = grpc.insecure_channel(server_uri)
                notification_stub = notification_service_pb2_grpc.NotificationServiceStub(channel)
                try:
                    request = ListMembersRequest(timeout_seconds=0)
                    response = notification_stub.listMembers(request)
                    if response.return_code == ReturnStatus.SUCCESS:
                        self.living_members = [proto_to_member(proto).server_uri
                                               for proto in response.members]
                    else:
                        raise Exception(response.return_msg)
                    self.current_uri = server_uri
                    self.notification_stub = notification_stub
                    break
                except grpc.RpcError as e:
                    last_error = e
            if self.current_uri is None:
                raise Exception("No available server uri!") from last_error
            self.ha_change_lock = threading.Lock()
            self.ha_running = True
            self.notification_stub = self._wrap_rpcs(self.notification_stub, server_uri)
            self.list_member_thread = threading.Thread(target=self._list_members, daemon=True)
            self.list_member_thread.start()

        if self._client_id is None:
            request = RegisterClientRequest(
                client_meta=ClientMeta(namespace=self.namespace, sender=self.sender))
            response = self.notification_stub.registerClient(request)
            if response.return_code == ReturnStatus.SUCCESS:
                self._client_id = response.client_id
            else:
                raise Exception(response.return_msg)
        else:
            request = ClientIdRequest(client_id=self._client_id)
            response = self.notification_stub.isClientExists(request)
            if response.return_code != ReturnStatus.SUCCESS:
                raise Exception("Failed to close notification client: {}".format(self))
            elif not response.is_exists:
                raise Exception("Init notification client with a client id which have not registered.")
        seq_num = 0 if self._initial_seq_num is None else int(self._initial_seq_num)
        self.sequence_num_manager = SequenceNumberManager(seq_num)

    def _ha_wrapper(self, func):
        @wraps(func)
        def call_with_retry(*args, **kwargs):
            current_func = getattr(self.notification_stub,
                                   func.__name__).inner_func
            start_time = time.time_ns() / 1000000
            failed_members = set()
            while True:
                try:
                    return current_func(*args, **kwargs)
                except grpc.RpcError:
                    logging.error("Exception thrown when calling rpc, change the connection.",
                                  exc_info=True)
                    with self.ha_change_lock:
                        # check the current_uri to ensure thread safety
                        if current_func.server_uri == self.current_uri:
                            living_members = list(self.living_members)
                            failed_members.add(self.current_uri)
                            shuffle(living_members)
                            found_new_member = False
                            for server_uri in living_members:
                                if server_uri in failed_members:
                                    continue
                                next_uri = server_uri
                                channel = grpc.insecure_channel(next_uri)
                                notification_stub = self._wrap_rpcs(
                                    notification_service_pb2_grpc.NotificationServiceStub(channel),
                                    next_uri)
                                self.notification_stub = notification_stub
                                current_func = getattr(self.notification_stub,
                                                       current_func.__name__).inner_func
                                self.current_uri = next_uri
                                found_new_member = True
                            if not found_new_member:
                                logging.error("No available living members currently. Sleep and retry.")
                                failed_members.clear()
                                sleep_and_detecting_running(self.retry_interval_ms,
                                                            lambda: self.ha_running)

                # break if stopped or timeout
                if not self.ha_running or \
                        time.time_ns() / 1000000 > start_time + self.retry_timeout_ms:
                    if not self.ha_running:
                        raise Exception("HA has been disabled.")
                    else:
                        raise Exception("Rpc retry timeout!")

        call_with_retry.inner_func = func
        return call_with_retry

    def _wrap_rpcs(self, stub, server_uri):
        for method_name, method in dict(stub.__dict__).items():
            method.__name__ = method_name
            method.server_uri = server_uri
            setattr(stub, method_name, self._ha_wrapper(method))
        return stub

    def _list_members(self):
        while self.ha_running:
            # refresh the living members
            request = ListMembersRequest(timeout_seconds=int(self.list_member_interval_ms / 1000))
            response = self.notification_stub.listMembers(request)
            if response.return_code == ReturnStatus.SUCCESS:
                with self.ha_change_lock:
                    self.living_members = [proto_to_member(proto).server_uri
                                           for proto in response.members]
            else:
                logging.error("Exception thrown when updating the living members: %s" %
                              response.return_msg)
            time.sleep(int(self.list_member_interval_ms / 1000))

    def disable_high_availability(self):
        if hasattr(self, "ha_running"):
            self.ha_running = False
            self.list_member_thread.join()

    def close(self):
        if self._client_id is not None:
            request = ClientIdRequest(client_id=self._client_id)
            response = self.notification_stub.deleteClient(request)
            if response.return_code != ReturnStatus.SUCCESS:
                raise Exception("Failed to close notification client: {}".format(self))
        logging.info("The notification client:{} has been closed.".format(self))

    @property
    def client_id(self):
        """Int. Id of the Client."""
        return self._client_id

    @property
    def sequence_number(self):
        return self.sequence_num_manager.get_sequence_number()

    def send_event(self, event: Event):
        """
        Send event to Notification Service.

        :param event: the event updated.
        :return: The created event which has offset and create time.
        """
        seq_num = self.sequence_num_manager.get_sequence_number()
        signature = '_'.join(['client', str(self._client_id), str(seq_num)])

        request = SendEventRequest(
            event=EventProto(
                key=event.key,
                value=event.value,
                context=event.context,
                namespace=self.namespace,
                sender=self.sender
            ),
            uuid=signature,
            enable_idempotence=True)
        response = self.notification_stub.sendEvent(request)
        if response.return_code == ReturnStatus.SUCCESS:
            self.sequence_num_manager.increment_sequence_number()
            return event_proto_to_event(response.event)
        else:
            raise Exception(response.return_msg)

    def register_listener(self,
                          listener_processor: ListenerProcessor,
                          event_keys: List[str] = None,
                          offset: int = None) -> ListenerRegistrationId:

        def list_events_from_offset(client, v, timeout_seconds: int = None):
            request = ListAllEventsRequest(start_offset=v, timeout_seconds=timeout_seconds)
            response = client.notification_stub.listAllEvents(request)
            if response.return_code == ReturnStatus.SUCCESS:
                if response.events is None:
                    return None
                else:
                    events = []
                    for event_proto in response.events:
                        event = event_proto_to_event(event_proto)
                        events.append(event)
                    return events
            else:
                raise Exception(response.return_msg)

        def listen(client, v, p):
            t = threading.current_thread()
            current_offset = 0 if v is None else v
            try:
                while getattr(t, '_flag', True):
                    notifications = list_events_from_offset(client,
                                                            current_offset,
                                                            NOTIFICATION_TIMEOUT_SECONDS)
                    if len(notifications) > 0:
                        if event_keys is not None:
                            events = client.filter_events(event_keys=event_keys, events=notifications)
                        else:
                            events = notifications
                        p.process(events)
                        current_offset = notifications[len(notifications) - 1].offset
            except Exception as e:
                logging.exception("Exception when listening events, %s", e)
                raise e

        thread = threading.Thread(target=listen,
                                  args=(self, offset, listener_processor),
                                  daemon=True)
        thread.start()
        self.lock.acquire()
        try:
            self.threads[thread.__hash__()] = thread
        finally:
            self.lock.release()
        return ListenerRegistrationId(id=str(thread.__hash__()))

    def unregister_listener(self, registration_id: ListenerRegistrationId):
        self.lock.acquire()
        key = int(registration_id.id)
        try:
            if key in self.threads:
                thread = self.threads[key]
                thread._flag = False
                thread.join()
                del self.threads[key]
        finally:
            self.lock.release()

    def list_all_events(self,
                        start_time: int = None,
                        start_offset: int = None,
                        end_offset: int = None) -> List[Event]:
        """
        List the range of `offset` of events in Notification Service.

        :param start_time: (Optional) Start time of the events.
        :param start_offset: (Optional) the offset of the events must greater than the
                              start_offset.
        :param end_offset: (Optional) the offset of the events must equal or less than the
                            end_offset.
        :return: The event list.
        """
        request = ListAllEventsRequest(start_time=start_time,
                                       start_offset=start_offset,
                                       end_offset=end_offset)
        response = self.notification_stub.listAllEvents(request)
        if response.return_code == ReturnStatus.SUCCESS:
            if response.events is None:
                return []
            else:
                events = []
                for event_proto in response.events:
                    event = event_proto_to_event(event_proto)
                    events.append(event)
                return events
        else:
            raise Exception(response.return_msg)

    def list_events(self,
                    key: str = None,
                    namespace: str = None,
                    sender: str = None,
                    begin_offset: int = None,
                    end_offset: int = None) -> List[Event]:
        """
        List specific events in Notification Service.

        :param key: Key of the event for listening.
        :param namespace: Namespace of the event for listening.
        :param sender: The event sender.
        :param begin_offset: Offset of the events must be greater than this offset.
        :param end_offset: Offset of the events must be less than or equal to this offset.
        :return: The event list.
        """
        request = ListEventsRequest(
            key=key,
            namespace=namespace,
            sender=sender,
            start_offset=begin_offset,
            end_offset=end_offset,
        )
        response = self.notification_stub.listEvents(request)
        if response.return_code == ReturnStatus.SUCCESS:
            if response.events is None:
                return []
            else:
                events = []
                for event_proto in response.events:
                    event = event_proto_to_event(event_proto)
                    events.append(event)
                return events
        else:
            raise Exception(response.return_msg)

    def time_to_offset(self, time: datetime) -> int:
        timestamp = int(time.timestamp() * 1000)
        request = TimeToOffsetRequest(timestamp=timestamp)
        response = self.notification_stub.timestampToEventOffset(request)
        return response.offset

    def count_events(self,
                     key: str = None,
                     namespace: str = None,
                     sender: str = None,
                     begin_offset: int = None,
                     end_offset: int = None) -> Tuple[int, List[SenderEventCount]]:
        """
        Count specific events in Notification Service.

        :param key: Key of the event for listening.
        :param namespace: Namespace of the event for listening.
        :param sender: The event sender.
        :param begin_offset: Offset of the events must be greater than this offset.
        :param end_offset: Offset of the events must be less than or equal to this offset.
        :return: The total event count and the list of event counts of each sender.
        """
        request = CountEventsRequest(
            key=key,
            namespace=namespace,
            sender=sender,
            start_offset=begin_offset,
            end_offset=end_offset,
        )
        response = self.notification_stub.countEvents(request)
        if response.return_code == ReturnStatus.SUCCESS:
            sender_event_counts = []
            for sender_event_count_proto in response.sender_event_counts:
                sender_event_count = event_count_proto_to_event_count(sender_event_count_proto)
                sender_event_counts.append(sender_event_count)
            return response.event_count, sender_event_counts
        else:
            raise Exception(response.return_msg)

    def filter_events(self, event_keys: List[str], events: List[Event]) -> List[Event]:
        def match(key, namespace, event: Event) -> bool:
            if key is not None and key != event.key:
                return False
            if namespace is not None and namespace != event.namespace:
                return False
            return True

        results = []
        for e in events:
            for k in event_keys:
                if match(k, self.namespace, e):
                    results.append(e)
                    break
        return results



# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Definition of event."""
from typing import Optional


DEFAULT_NAMESPACE = "DEFAULT"
ANY_CONDITION = "*"


class Event(object):
    def __init__(self,
                 key: str,
                 value: Optional[str],
                 ):
        """
        Event represents an event.
        :param key: It represents the key of the event.
        :param value: It represents the value of the event.
        """
        self.key = key
        self.value = value

        self._namespace = None
        self._sender = None
        self._create_time = None
        self._offset = None
        self._context = None

    @property
    def namespace(self):
        return self._namespace

    @namespace.setter
    def namespace(self, value):
        self._namespace = value

    @property
    def sender(self):
        return self._sender

    @sender.setter
    def sender(self, value):
        self._sender = value

    @property
    def context(self):
        return self._context

    @context.setter
    def context(self, value):
        self._context = value

    @property
    def create_time(self):
        return self._create_time

    @create_time.setter
    def create_time(self, value):
        self._create_time = value

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, value):
        self._offset = value

    def __str__(self) -> str:
        return 'key:{0}, value:{1}, namespace:{2}, sender:{3}, create_time:{4}, offset:{5}, context: {6}' \
            .format(self.key,
                    self.value,
                    self._namespace,
                    self._sender,
                    self._create_time,
                    self._offset,
                    self._context)

    def __eq__(self, other):
        if not isinstance(other, Event):
            return False
        return self.key == other.key \
               and self.value == other.value \
               and self._namespace == other.namespace \
               and self.sender == other.sender \
               and self._context == other.context \
               and self._create_time == other._create_time \
               and self._offset == other._offset



#!/usr/bin/env python
#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Command-line interface"""

import argparse
from argparse import Action, RawTextHelpFormatter
from datetime import datetime
from functools import lru_cache
from typing import Callable, Dict, Iterable, List, NamedTuple, Optional, Union

from notification_service.util.cli import ColorMode
from notification_service.util.utils import import_string, partition, parse_date


def lazy_load_command(import_path: str) -> Callable:
    """Create a lazy loader for command"""
    _, _, name = import_path.rpartition('.')

    def command(*args, **kwargs):
        func = import_string(import_path)
        return func(*args, **kwargs)

    command.__name__ = name

    return command


class DefaultHelpParser(argparse.ArgumentParser):
    """CustomParser to display help message"""

    def error(self, message):
        """Override error and use print_instead of print_usage"""
        self.print_help()
        self.exit(2, f'\n{self.prog} command error: {message}, see help above.\n')


class Arg:
    """Class to keep information about command line argument"""
    # Used in Arg to enable `None' as a distinct value from "not passed"
    _UNSET = object()

    # pylint: disable=redefined-builtin,unused-argument,too-many-arguments
    def __init__(
            self,
            flags=_UNSET,
            help=_UNSET,
            action=_UNSET,
            default=_UNSET,
            nargs=_UNSET,
            type=_UNSET,
            choices=_UNSET,
            required=_UNSET,
            metavar=_UNSET,
            dest=_UNSET,
    ):
        self.flags = flags
        self.kwargs = {}
        for k, v in locals().items():
            if v is Arg._UNSET:
                continue
            if k in ("self", "flags"):
                continue

            self.kwargs[k] = v

    # pylint: enable=redefined-builtin,unused-argument,too-many-arguments

    def add_to_parser(self, parser: argparse.ArgumentParser):
        """Add this argument to an ArgumentParser"""
        parser.add_argument(*self.flags, **self.kwargs)


class ActionCommand(NamedTuple):
    """Single CLI command"""

    name: str
    help: str
    func: Callable
    args: Iterable[Arg]
    description: Optional[str] = None
    epilog: Optional[str] = None


class GroupCommand(NamedTuple):
    """ClI command with subcommands"""

    name: str
    help: str
    subcommands: Iterable
    description: Optional[str] = None
    epilog: Optional[str] = None


ARG_YES = Arg(
    ("-y", "--yes"), help="Do not prompt to confirm reset. Use with care!", action="store_true", default=False
)

ARG_DB_VERSION = Arg(
    ("-v", "--version"),
    help=(
        'The version corresponding to the database.'
    ),
    default='heads',
)

ARG_SERVER_DAEMON = Arg(
    ("-d", "--daemon"),
    help="Daemonize instead of running in the foreground",
    action="store_true"
)

CLICommand = Union[ActionCommand, GroupCommand]

ARG_SERVER_URI = Arg(
    ("-s", "--server-uri"),
    help="The uri of notification server",
)

ARG_EVENT_NAMESPACE = Arg(
    ("-n", "--namespace"),
    help="Namespace of the event. If not set, all namespaces would be handled",
)

ARG_EVENT_KEY = Arg(
    ("key",),
    help="Key of the event",
)

ARG_EVENT_VALUE = Arg(
    ("value",),
    help="Value of the event",
)

ARG_EVENT_CONTEXT = Arg(
    ("--context",),
    help="Context of the event",
)

ARG_BEGIN_OFFSET = Arg(
    ("--begin-offset",),
    help="Begin offset of the event. Defaults to 0",
    type=int,
    default=0,
)

ARG_BEGIN_TIME = Arg(
    ("--begin-time",),
    help="Begin datetime of the event, formatted in ISO 8601",
    type=parse_date,
)

ARG_LISTEN_BEGIN_TIME = Arg(
    ("--begin-time",),
    help="Begin datetime of the event to listen, formatted in ISO 8601",
    type=parse_date,
    default=datetime.now().isoformat()
)

ARG_EVENT_SENDER = Arg(
    ("--sender",),
    help="Sender of the event",
)

# config
ARG_OPTION = Arg(
    ("option",),
    help="The option name of the configuration",
)

ARG_OUTPUT = Arg(
    ("-o", "--output"),
    help="Output format. Allowed values: json, yaml, table (default: table)",
    metavar="(table, json, yaml)",
    choices=("table", "json", "yaml"),
    default="table",
)

ARG_COLOR = Arg(
    ('--color',),
    help="Does emit colored config output (default: auto)",
    choices={ColorMode.ON, ColorMode.OFF, ColorMode.AUTO},
    default=ColorMode.AUTO,
)

ARG_TIMEOUT = Arg(
    ("--timeout",),
    help="Client timeout in seconds.",
    default=None,
    type=int,
)

EVENT_COMMANDS = (
    ActionCommand(
        name='list',
        help="Lists events",
        func=lazy_load_command('notification_service.cli.commands.event_command.list_events'),
        args=(ARG_EVENT_KEY, ARG_SERVER_URI, ARG_EVENT_NAMESPACE, ARG_BEGIN_OFFSET,
              ARG_BEGIN_TIME, ARG_EVENT_SENDER, ARG_OUTPUT),
    ),
    ActionCommand(
        name='count',
        help='Counts events',
        func=lazy_load_command('notification_service.cli.commands.event_command.count_events'),
        args=(ARG_EVENT_KEY, ARG_SERVER_URI, ARG_EVENT_NAMESPACE, ARG_BEGIN_OFFSET,
              ARG_BEGIN_TIME, ARG_EVENT_SENDER),
    ),
    ActionCommand(
        name='listen',
        help='Listens events',
        func=lazy_load_command('notification_service.cli.commands.event_command.listen_events'),
        args=(ARG_EVENT_KEY, ARG_SERVER_URI, ARG_EVENT_NAMESPACE, ARG_BEGIN_OFFSET,
              ARG_LISTEN_BEGIN_TIME)
    ),
    ActionCommand(
        name='send',
        help='Sends an event',
        func=lazy_load_command('notification_service.cli.commands.event_command.send_event'),
        args=(ARG_EVENT_KEY, ARG_EVENT_VALUE, ARG_SERVER_URI, ARG_EVENT_NAMESPACE,
              ARG_EVENT_CONTEXT, ARG_EVENT_SENDER)
    )
)

DB_COMMANDS = (
    ActionCommand(
        name='init',
        help="Initializes the metadata database",
        func=lazy_load_command('notification_service.cli.commands.db_command.init'),
        args=(),
    ),
    ActionCommand(
        name='reset',
        help="Burns down and rebuild the metadata database",
        func=lazy_load_command('notification_service.cli.commands.db_command.reset'),
        args=(ARG_YES,),
    ),
    ActionCommand(
        name='upgrade',
        help="Upgrades the metadata database to the version",
        func=lazy_load_command('notification_service.cli.commands.db_command.upgrade'),
        args=(ARG_DB_VERSION,),
    ),
    ActionCommand(
        name='downgrade',
        help="Downgrades the metadata database to the version",
        func=lazy_load_command('notification_service.cli.commands.db_command.downgrade'),
        args=(ARG_DB_VERSION,),
    )
)

SERVER_COMMANDS = (
    ActionCommand("start",
                  "Starts the notification server",
                  lazy_load_command("notification_service.cli.commands.server_command.server_start"),
                  [ARG_SERVER_DAEMON],
                  "Start the notification server"),

    ActionCommand("stop",
                  "Stops the notification server",
                  lazy_load_command("notification_service.cli.commands.server_command.server_stop"),
                  [],
                  "Stop the notification server")
)

CONFIG_COMMANDS = (
    ActionCommand(
        name='get-value',
        help='Gets the option value of the configuration.',
        func=lazy_load_command('notification_service.cli.commands.config_command.config_get_value'),
        args=(ARG_OPTION,),
    ),
    ActionCommand(
        name='init',
        help='Initializes the default configuration.',
        func=lazy_load_command('notification_service.cli.commands.config_command.config_init'),
        args=(),
    ),
    ActionCommand(
        name='list',
        help='List all options of the configuration.',
        func=lazy_load_command('notification_service.cli.commands.config_command.config_list'),
        args=(ARG_COLOR,),
    ),
)

notification_commands: List[CLICommand] = [
    ActionCommand("version",
                  "Shows the version of Notification Service",
                  lazy_load_command("notification_service.cli.commands.version_command.version"),
                  [],
                  "Shows the version of Notification Service"),
    GroupCommand(
        name='server',
        help='Notification server operations',
        subcommands=SERVER_COMMANDS
    ),
    GroupCommand(
        name='event',
        help='Manage events',
        subcommands=EVENT_COMMANDS,
    ),
    GroupCommand(
        name='db',
        help="Database operations",
        subcommands=DB_COMMANDS,
    ),
    GroupCommand(
        name="config",
        help='Manage configuration',
        subcommands=CONFIG_COMMANDS
    ),
]
ALL_COMMANDS_DICT: Dict[str, CLICommand] = {sp.name: sp for sp in notification_commands}


class NotificationHelpFormatter(argparse.HelpFormatter):
    """
    Custom help formatter to display help message.

    It displays simple commands and groups of commands in separate sections.
    """

    def _format_action(self, action: Action):
        if isinstance(action, argparse._SubParsersAction):  # pylint: disable=protected-access

            parts = []
            action_header = self._format_action_invocation(action)
            action_header = '%*s%s\n' % (self._current_indent, '', action_header)
            parts.append(action_header)

            self._indent()
            subactions = action._get_subactions()  # pylint: disable=protected-access
            action_subcommands, group_subcommands = partition(
                lambda d: isinstance(ALL_COMMANDS_DICT[d.dest], GroupCommand), subactions
            )
            parts.append("\n")
            parts.append('%*s%s:\n' % (self._current_indent, '', "Groups"))
            self._indent()
            for subaction in group_subcommands:
                parts.append(self._format_action(subaction))
            self._dedent()

            parts.append("\n")
            parts.append('%*s%s:\n' % (self._current_indent, '', "Commands"))
            self._indent()

            for subaction in action_subcommands:
                parts.append(self._format_action(subaction))
            self._dedent()
            self._dedent()

            # return a single string
            return self._join_parts(parts)

        return super()._format_action(action)


@lru_cache(maxsize=None)
def get_parser() -> argparse.ArgumentParser:
    """Creates and returns command line argument parser"""
    parser = DefaultHelpParser(prog="notification", formatter_class=NotificationHelpFormatter)
    subparsers = parser.add_subparsers(dest='subcommand', metavar="GROUP_OR_COMMAND")
    subparsers.required = True

    sub_name: str
    for sub_name in sorted(ALL_COMMANDS_DICT.keys()):
        sub: CLICommand = ALL_COMMANDS_DICT[sub_name]
        _add_command(subparsers, sub)
    return parser


def _sort_args(args: Iterable[Arg]) -> Iterable[Arg]:
    """Sort subcommand optional args, keep positional args"""

    def get_long_option(arg: Arg):
        """Get long option from Arg.flags"""
        return arg.flags[0] if len(arg.flags) == 1 else arg.flags[1]

    positional, optional = partition(lambda x: x.flags[0].startswith("-"), args)
    yield from positional
    yield from sorted(optional, key=lambda x: get_long_option(x).lower())


def _add_command(
        subparsers: argparse._SubParsersAction, sub: CLICommand  # pylint: disable=protected-access
) -> None:
    sub_proc = subparsers.add_parser(
        sub.name, help=sub.help, description=sub.description or sub.help, epilog=sub.epilog
    )
    sub_proc.formatter_class = RawTextHelpFormatter

    if isinstance(sub, GroupCommand):
        _add_group_command(sub, sub_proc)
    elif isinstance(sub, ActionCommand):
        _add_action_command(sub, sub_proc)
    else:
        raise ValueError("Invalid command definition.")


def _add_action_command(sub: ActionCommand, sub_proc: argparse.ArgumentParser) -> None:
    for arg in _sort_args(sub.args):
        arg.add_to_parser(sub_proc)
    sub_proc.set_defaults(func=sub.func)


def _add_group_command(sub: GroupCommand, sub_proc: argparse.ArgumentParser) -> None:
    subcommands = sub.subcommands
    sub_subparsers = sub_proc.add_subparsers(dest="subcommand", metavar="COMMAND")
    sub_subparsers.required = True

    for command in sorted(subcommands, key=lambda x: x.name):
        _add_command(sub_subparsers, command)



#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#


class SenderEventCount(object):
    def __init__(self, sender: str, event_count: int):
        self.sender = sender
        self.event_count = event_count

    def __str__(self) -> str:
        return 'sender:{0}, event_count:{1}'.format(self.sender, self.event_count)

    def __eq__(self, other):
        if not isinstance(other, SenderEventCount):
            return False
        return self.sender == other.sender and self.event_count == other.event_count



#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import logging
import time
from collections import Iterable
from typing import Tuple, Union

from notification_service.storage.alchemy.db_client_storage import ClientModel
from sqlalchemy import Column, String, BigInteger, Text, Integer, func

from notification_service.model.event import Event, ANY_CONDITION
from notification_service.storage.event_storage import BaseEventStorage
from notification_service.util import db
from notification_service.util.db import provide_session
from notification_service.util.utils import event_model_to_event, count_result_to_sender_event_count

from notification_service.storage.alchemy.base import Base


class DbEventStorage(BaseEventStorage):

    def __init__(self, db_conn=None):
        if db_conn is not None:
            db.SQL_ALCHEMY_CONN = db_conn
        db.prepare_db()

    def add_event(self, event: Event, uuid: str):
        return EventModel.add_event(event, uuid)

    def list_events(self,
                    key: str = None,
                    namespace: str = None,
                    sender: str = None,
                    begin_offset: int = None,
                    end_offset: int = None):
        return EventModel.list_events(key, namespace, sender, begin_offset, end_offset)

    def count_events(self,
                     key: str = None,
                     namespace: str = None,
                     sender: str = None,
                     begin_offset: int = None,
                     end_offset: int = None):
        return EventModel.count_events(key, namespace, sender, begin_offset, end_offset)

    def list_all_events(self, start_time: int):
        return EventModel.list_all_events(start_time)

    def list_all_events_from_offset(self, start_offset: int, end_offset: int = None):
        return EventModel.list_all_events_from_offset(start_offset, end_offset)

    def register_client(self, namespace: str = None, sender: str = None) -> int:
        return ClientModel.register_client(namespace, sender)

    def delete_client(self, client_id):
        ClientModel.delete_client(client_id)

    def is_client_exists(self, client_id) -> bool:
        return ClientModel.is_client_exists(client_id)

    def get_event_by_uuid(self, uuid: str):
        return EventModel.get_event_by_uuid(uuid)

    def timestamp_to_event_offset(self, timestamp: int) -> int:
        return EventModel.timestamp_to_offset(timestamp)

    def clean_up(self):
        tables = set(db.engine.table_names())
        for tbl in reversed(Base.metadata.sorted_tables):
            if tbl.name in tables:
                logging.info('Delete all data from table: {}', tbl.name)
                db.engine.execute(tbl.delete())


class EventModel(Base):
    __tablename__ = "event_model"
    offset = Column(BigInteger().with_variant(Integer, "sqlite"), primary_key=True)
    key = Column(String(1024), nullable=False)
    value = Column(Text())
    context = Column(Text())
    namespace = Column(String(1024))
    sender = Column(String(1024))
    create_time = Column(BigInteger(), nullable=False)
    uuid = Column(String(40), nullable=False, unique=True)

    @staticmethod
    @provide_session
    def add_event(event: Event, uuid, session=None):
        event_model = EventModel()
        event_model.key = event.key
        event_model.value = event.value
        event_model.context = event.context
        event_model.namespace = event.namespace
        event_model.create_time = int(time.time() * 1000)
        event_model.uuid = uuid
        event_model.sender = event.sender
        session.add(event_model)
        session.commit()
        return event_model_to_event(event_model)

    @staticmethod
    @provide_session
    def list_events(key: str = None,
                    namespace: str = None,
                    sender: str = None,
                    begin_offset: int = None,
                    end_offset: int = None,
                    session=None):
        key = None if key == "" else key
        namespace = None if namespace == "" else namespace
        sender = None if sender == "" else sender

        conditions = []
        if key is not None and ANY_CONDITION != key:
            conditions.append(EventModel.key == key)
        if namespace is not None and ANY_CONDITION != namespace:
            conditions.append(EventModel.namespace == namespace)
        if sender is not None and ANY_CONDITION != sender:
            conditions.append(EventModel.sender == sender)
        if begin_offset and begin_offset > 0:
            conditions.append(EventModel.offset > begin_offset)
        if end_offset and end_offset > 0:
            conditions.append(EventModel.offset <= begin_offset)
        event_model_list = session.query(EventModel).filter(*conditions).all()
        return [event_model_to_event(event_model) for event_model in event_model_list]

    @staticmethod
    @provide_session
    def count_events(key: str = None,
                     namespace: str = None,
                     sender: str = None,
                     begin_offset: int = None,
                     end_offset: int = None,
                     session=None):
        key = None if key == "" else key
        namespace = None if namespace == "" else namespace
        sender = None if sender == "" else sender
        conditions = []
        if key is not None and ANY_CONDITION != key:
            conditions.append(EventModel.key == key)
        if namespace is not None and ANY_CONDITION != namespace:
            conditions.append(EventModel.namespace == namespace)
        if sender is not None and ANY_CONDITION != sender:
            conditions.append(EventModel.sender == sender)
        if begin_offset and begin_offset > 0:
            conditions.append(EventModel.offset > begin_offset)
        if end_offset and end_offset > 0:
            conditions.append(EventModel.offset <= begin_offset)

        count_results = session.query(EventModel.sender, func.count('*').label('event_count'))\
            .filter(*conditions).group_by(EventModel.sender)
        return [count_result_to_sender_event_count(count_result) for count_result in count_results]

    @staticmethod
    @provide_session
    def list_all_events(start_time: int, session=None):
        conditions = [
            EventModel.create_time >= start_time
        ]
        event_model_list = session.query(EventModel).filter(*conditions).all()
        return [event_model_to_event(event_model) for event_model in event_model_list]

    @staticmethod
    @provide_session
    def list_all_events_from_offset(start_offset: int, end_offset: int = None, session=None):
        conditions = [
            EventModel.offset > start_offset
        ]
        if end_offset is not None and end_offset > 0:
            conditions.append(EventModel.offset <= end_offset)
        event_model_list = session.query(EventModel).filter(*conditions).all()
        return [event_model_to_event(event_model) for event_model in event_model_list]

    @staticmethod
    @provide_session
    def timestamp_to_offset(timestamp, session=None):
        e = session.query(EventModel).filter(EventModel.create_time <= timestamp) \
            .order_by(EventModel.offset.desc()) \
            .limit(1).first()
        return e.offset if e else 0

    @staticmethod
    @provide_session
    def get_event_by_uuid(uuid, session=None):
        conditions = [
            EventModel.uuid == uuid
        ]
        return session.query(EventModel).filter(*conditions).first()



# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import yaml

import logging
import os

from typing import Dict

logger = logging.getLogger(__name__)


class NotificationServerConfig(object):
    def __init__(self, config_file):
        self.config_file = config_file
        self._port = 50052
        self._db_uri = None
        self._enable_ha = False
        self._ha_ttl_ms = None
        self._advertised_uri = None
        self._wait_for_server_started_timeout = 5.0
        self._parse_config()

    def _parse_config(self):
        with open(self.config_file, 'r') as f:
            yaml_config = yaml.load(f, Loader=yaml.FullLoader)
            if 'server_port' in yaml_config:
                self._port = yaml_config['server_port']
            if 'db_uri' in yaml_config:
                self._db_uri = yaml_config['db_uri']
            if 'enable_ha' in yaml_config:
                self._enable_ha = yaml_config['enable_ha']
            if 'ha_ttl_ms' in yaml_config:
                self._ha_ttl_ms = yaml_config['ha_ttl_ms']
            if 'advertised_uri' in yaml_config:
                self._advertised_uri = yaml_config['advertised_uri']
            if 'wait_for_server_started_timeout' in yaml_config:
                self._wait_for_server_started_timeout = yaml_config['wait_for_server_started_timeout']

    @property
    def port(self):
        return self._port

    @port.setter
    def port(self, port):
        self._port = port

    @property
    def db_uri(self):
        return self._db_uri

    @db_uri.setter
    def db_uri(self, db_uri):
        self._db_uri = db_uri

    @property
    def enable_ha(self):
        return self._enable_ha

    @enable_ha.setter
    def enable_ha(self, enable_ha):
        self._enable_ha = enable_ha

    @property
    def ha_ttl_ms(self):
        return self._ha_ttl_ms

    @ha_ttl_ms.setter
    def ha_ttl_ms(self, ha_ttl_ms):
        self._ha_ttl_ms = ha_ttl_ms

    @property
    def advertised_uri(self):
        return self._advertised_uri

    @advertised_uri.setter
    def advertised_uri(self, advertised_uri):
        self._advertised_uri = advertised_uri

    @property
    def wait_for_server_started_timeout(self):
        return self._wait_for_server_started_timeout

    @wait_for_server_started_timeout.setter
    def wait_for_server_started_timeout(self, wait_for_server_started_timeout):
        self._wait_for_server_started_timeout = wait_for_server_started_timeout

    def __repr__(self):
        return f"""
port: {self.port}
db_uri: {self.db_uri}
        """


def create_server_config(root_dir_path, param: Dict[str, str]):
    """
    Generate the default config of the Notification Server.
    """
    import notification_service.config_templates
    default_config_file_name = 'default_notification_server.yaml'
    config_file_name = 'notification_server.yaml'

    if not os.path.exists(root_dir_path):
        logging.info('{} does not exist, creating the directory.'.format(root_dir_path))
        os.makedirs(root_dir_path, exist_ok=False)
    default_server_config_path = os.path.join(
        os.path.dirname(notification_service.config_templates.__file__), default_config_file_name
    )
    if not os.path.exists(default_server_config_path):
        raise Exception('Default notification server config is not found at {}.'.format(default_server_config_path))

    server_config_target_path = os.path.join(root_dir_path, config_file_name)
    with open(default_server_config_path, encoding='utf-8') as config_file:
        default_config = config_file.read().format(**param)
    with open(server_config_target_path, mode='w', encoding='utf-8') as f:
        f.write(default_config)
    return server_config_target_path



# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Server command"""
import datetime
import logging
import os
import signal

import daemon
from daemon.pidfile import TimeoutPIDLockFile

import notification_service.settings
from notification_service.server.server_runner import NotificationServerRunner
from notification_service.settings import get_configuration_file_path
from notification_service.util.utils import check_pid_exist, stop_process

logger = logging.getLogger(__name__)


def sigterm_handler(signum, frame):
    # We raise KeyboardInterrupt so that the server can stop gracefully
    raise KeyboardInterrupt()


def make_log_dir_if_not_exist():
    log_dir = os.path.join(notification_service.settings.NOTIFICATION_HOME, "logs")
    if os.path.exists(log_dir):
        return

    os.makedirs(log_dir, exist_ok=True)


def server_start(args):
    pid_file_path = os.path.join(notification_service.settings.NOTIFICATION_HOME,
                                 notification_service.settings.NOTIFICATION_PID_FILENAME)
    if os.path.exists(pid_file_path):
        with open(pid_file_path, 'r') as f:
            pid = int(f.read())
        if not check_pid_exist(pid):
            logger.info(
                "Process pid: {} does not exist. This means a staled PID file. Removing the PID file".format(pid))
            os.remove(pid_file_path)
        else:
            logger.info("Notification Server is running, stop it first with 'notification server stop'.")
            return
    if args.daemon:
        make_log_dir_if_not_exist()
        log_path = os.path.join(notification_service.settings.NOTIFICATION_HOME, "logs",
                                'notification_server-{}.log'.format(datetime.datetime.now().strftime("%Y%m%d-%H%M%S")))

        logger.info(f"\nStarting Notification Server in daemon mode\n"
                    f"Notification server log: {log_path}\n"
                    f"Notification server pid file: {pid_file_path}")

        log = open(log_path, 'w+')
        ctx = _get_daemon_context(log, pid_file_path)
        with ctx:
            config_file_path = get_configuration_file_path()
            server_runner = NotificationServerRunner(config_file_path)
            server_runner.start(True)

        log.close()
    else:
        """Start the notification server"""
        signal.signal(signal.SIGTERM, sigterm_handler)
        config_file_path = get_configuration_file_path()

        pid = os.getpid()
        logger.info("Starting notification server at pid: {}".format(pid))
        with open(pid_file_path, 'w') as f:
            f.write(str(pid))

        try:
            server_runner = NotificationServerRunner(config_file_path)
            server_runner.start(True)
        finally:
            if os.path.exists(pid_file_path):
                os.remove(pid_file_path)


def server_stop(args):
    pid_file_path = os.path.join(notification_service.settings.NOTIFICATION_HOME,
                                 notification_service.settings.NOTIFICATION_PID_FILENAME)
    if not os.path.exists(pid_file_path):
        logger.info(
            "PID file of Notification server does not exist at {}. The Notification server is not running.".format(
                pid_file_path))
        return

    with open(pid_file_path, 'r') as f:
        pid = int(f.read())

    if not check_pid_exist(pid):
        logger.info("Process pid: {} does not exist. This means a staled PID file. Removing the PID file".format(pid))
        os.remove(pid_file_path)
        return

    stop_process(pid, "Notification server")


def _get_daemon_context(log, pid_file_path: str) -> daemon.DaemonContext:
    return daemon.DaemonContext(
        pidfile=TimeoutPIDLockFile(pid_file_path, -1),
        stdout=log,
        stderr=log,
        signal_map={
            signal.SIGTERM: sigterm_handler
        }
    )



# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""add_member

Revision ID: 29623a4151ed
Revises: 87cb292bcc31
Create Date: 2021-10-15 12:06:45.611155

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '29623a4151ed'
down_revision = '87cb292bcc31'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('member_model',
                    sa.Column('id', sa.BigInteger().with_variant(sa.Integer(), 'sqlite'), nullable=False),
                    sa.Column('version', sa.BigInteger(), nullable=False),
                    sa.Column('server_uri', sa.String(length=767), nullable=False),
                    sa.Column('update_time', sa.BigInteger(), nullable=False),
                    sa.Column('uuid', sa.String(length=128), nullable=False),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('server_uri'),
                    sa.UniqueConstraint('uuid')
                    )


def downgrade():
    op.drop_table('member_model')


# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Config command"""
import logging
import os

import pygments
import yaml
from pygments.lexers.configs import IniLexer

from notification_service.settings import get_configuration_file_path, NOTIFICATION_HOME
from notification_service.util.cli import should_use_colors, get_terminal_formatter
from notification_service.util.server_config import create_server_config

logger = logging.getLogger(__name__)


def config_get_value(args):
    """Gets the option value of the configuration."""
    with open(get_configuration_file_path(), 'r') as f:
        yaml_config = yaml.load(f, Loader=yaml.FullLoader)
        if args.option not in yaml_config:
            raise SystemExit(f'The option [{args.option}] is not found in config.')
        value = yaml_config[args.option]
        print(value)


def config_init(args):
    """Initializes the default configuration."""
    if not os.path.exists(NOTIFICATION_HOME):
        os.makedirs(NOTIFICATION_HOME)
    config_file = NOTIFICATION_HOME + '/notification_server.yaml'
    if os.path.exists(config_file):
        logger.info('Notification server config has already initialized at {}.'.format(config_file))
    else:
        create_server_config(NOTIFICATION_HOME, {'NOTIFICATION_HOME': NOTIFICATION_HOME})
        logger.info('Notification server config generated at {}.'.format(config_file))


def config_list(args):
    """List all options of the configuration."""
    with open(get_configuration_file_path(), 'r') as f:
        code = f.read()
        if should_use_colors(args):
            code = pygments.highlight(code=code, formatter=get_terminal_formatter(), lexer=IniLexer())
        print(code)



#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: notification_service.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import service as _service
from google.protobuf import service_reflection
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='notification_service.proto',
  package='notification_service',
  syntax='proto3',
  serialized_options=b'\n\035org.aiflow.notification.protoZ\025/notification_service\210\001\001\220\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1anotification_service.proto\x12\x14notification_service\x1a\x1cgoogle/api/annotations.proto\"\x81\x01\n\nEventProto\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\x12\x0f\n\x07\x63ontext\x18\x03 \x01(\t\x12\x11\n\tnamespace\x18\x04 \x01(\t\x12\x0e\n\x06offset\x18\x05 \x01(\x03\x12\x13\n\x0b\x63reate_time\x18\x06 \x01(\x03\x12\x0e\n\x06sender\x18\x07 \x01(\t\"<\n\x15SenderEventCountProto\x12\x0e\n\x06sender\x18\x01 \x01(\t\x12\x13\n\x0b\x65vent_count\x18\x02 \x01(\x03\"m\n\x10SendEventRequest\x12/\n\x05\x65vent\x18\x01 \x01(\x0b\x32 .notification_service.EventProto\x12\x0c\n\x04uuid\x18\x02 \x01(\t\x12\x1a\n\x12\x65nable_idempotence\x18\x03 \x01(\x08\"/\n\nClientMeta\x12\x11\n\tnamespace\x18\x01 \x01(\t\x12\x0e\n\x06sender\x18\x02 \x01(\t\"\x92\x01\n\x12SendEventsResponse\x12/\n\x05\x65vent\x18\x01 \x01(\x0b\x32 .notification_service.EventProto\x12\x37\n\x0breturn_code\x18\x02 \x01(\x0e\x32\".notification_service.ReturnStatus\x12\x12\n\nreturn_msg\x18\x03 \x01(\t\"\x86\x01\n\x11ListEventsRequest\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x11\n\tnamespace\x18\x02 \x01(\t\x12\x0e\n\x06sender\x18\x03 \x01(\t\x12\x14\n\x0cstart_offset\x18\x04 \x01(\x03\x12\x12\n\nend_offset\x18\x05 \x01(\x03\x12\x17\n\x0ftimeout_seconds\x18\x06 \x01(\x05\"n\n\x12\x43ountEventsRequest\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x11\n\tnamespace\x18\x02 \x01(\t\x12\x0e\n\x06sender\x18\x03 \x01(\t\x12\x14\n\x0cstart_offset\x18\x04 \x01(\x03\x12\x12\n\nend_offset\x18\x05 \x01(\x03\"m\n\x14ListAllEventsRequest\x12\x17\n\x0ftimeout_seconds\x18\x01 \x01(\x05\x12\x12\n\nstart_time\x18\x02 \x01(\x03\x12\x14\n\x0cstart_offset\x18\x03 \x01(\x03\x12\x12\n\nend_offset\x18\x04 \x01(\x03\"\x93\x01\n\x12ListEventsResponse\x12\x37\n\x0breturn_code\x18\x01 \x01(\x0e\x32\".notification_service.ReturnStatus\x12\x12\n\nreturn_msg\x18\x02 \x01(\t\x12\x30\n\x06\x65vents\x18\x03 \x03(\x0b\x32 .notification_service.EventProto\"\xc1\x01\n\x13\x43ountEventsResponse\x12\x37\n\x0breturn_code\x18\x01 \x01(\x0e\x32\".notification_service.ReturnStatus\x12\x12\n\nreturn_msg\x18\x02 \x01(\t\x12\x13\n\x0b\x65vent_count\x18\x03 \x01(\x03\x12H\n\x13sender_event_counts\x18\x04 \x03(\x0b\x32+.notification_service.SenderEventCountProto\"G\n\x0bMemberProto\x12\x0f\n\x07version\x18\x01 \x01(\x03\x12\x12\n\nserver_uri\x18\x02 \x01(\t\x12\x13\n\x0bupdate_time\x18\x03 \x01(\x03\"(\n\x06Notify\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x11\n\tnamespace\x18\x02 \x01(\t\"?\n\rNotifyRequest\x12.\n\x08notifies\x18\x01 \x03(\x0b\x32\x1c.notification_service.Notify\"-\n\x12ListMembersRequest\x12\x17\n\x0ftimeout_seconds\x18\x01 \x01(\x05\"\x96\x01\n\x13ListMembersResponse\x12\x37\n\x0breturn_code\x18\x01 \x01(\x0e\x32\".notification_service.ReturnStatus\x12\x12\n\nreturn_msg\x18\x02 \x01(\t\x12\x32\n\x07members\x18\x03 \x03(\x0b\x32!.notification_service.MemberProto\"K\n\x16NotifyNewMemberRequest\x12\x31\n\x06member\x18\x01 \x01(\x0b\x32!.notification_service.MemberProto\"N\n\x15RegisterClientRequest\x12\x35\n\x0b\x63lient_meta\x18\x01 \x01(\x0b\x32 .notification_service.ClientMeta\"x\n\x16RegisterClientResponse\x12\x37\n\x0breturn_code\x18\x01 \x01(\x0e\x32\".notification_service.ReturnStatus\x12\x12\n\nreturn_msg\x18\x02 \x01(\t\x12\x11\n\tclient_id\x18\x03 \x01(\x03\"$\n\x0f\x43lientIdRequest\x12\x11\n\tclient_id\x18\x01 \x01(\x03\"]\n\x0e\x43ommonResponse\x12\x37\n\x0breturn_code\x18\x01 \x01(\x0e\x32\".notification_service.ReturnStatus\x12\x12\n\nreturn_msg\x18\x02 \x01(\t\"x\n\x16isClientExistsResponse\x12\x37\n\x0breturn_code\x18\x01 \x01(\x0e\x32\".notification_service.ReturnStatus\x12\x12\n\nreturn_msg\x18\x02 \x01(\t\x12\x11\n\tis_exists\x18\x03 \x01(\x08\"(\n\x13TimeToOffsetRequest\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\"_\n\x14TimeToOffsetResponse\x12\x37\n\x0breturn_code\x18\x01 \x01(\x0e\x32\".notification_service.ReturnStatus\x12\x0e\n\x06offset\x18\x02 \x01(\x03*&\n\x0cReturnStatus\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\x32\xc7\r\n\x13NotificationService\x12\x91\x01\n\tsendEvent\x12&.notification_service.SendEventRequest\x1a(.notification_service.SendEventsResponse\"2\x82\xd3\xe4\x93\x02,\"\'/aiflow/notification-service/send_event:\x01*\x12\x91\x01\n\nlistEvents\x12\'.notification_service.ListEventsRequest\x1a(.notification_service.ListEventsResponse\"0\x82\xd3\xe4\x93\x02*\x12(/aiflow/notification-service/list_events\x12\x95\x01\n\x0b\x63ountEvents\x12(.notification_service.CountEventsRequest\x1a).notification_service.CountEventsResponse\"1\x82\xd3\xe4\x93\x02+\x12)/aiflow/notification-service/count_events\x12\x9b\x01\n\rlistAllEvents\x12*.notification_service.ListAllEventsRequest\x1a(.notification_service.ListEventsResponse\"4\x82\xd3\xe4\x93\x02.\x12,/aiflow/notification-service/list_all_events\x12\x83\x01\n\x06notify\x12#.notification_service.NotifyRequest\x1a$.notification_service.CommonResponse\".\x82\xd3\xe4\x93\x02(\"#/aiflow/notification-service/notify:\x01*\x12\x95\x01\n\x0blistMembers\x12(.notification_service.ListMembersRequest\x1a).notification_service.ListMembersResponse\"1\x82\xd3\xe4\x93\x02+\x12)/aiflow/notification-service/list_members\x12\xa0\x01\n\x0fnotifyNewMember\x12,.notification_service.NotifyNewMemberRequest\x1a$.notification_service.CommonResponse\"9\x82\xd3\xe4\x93\x02\x33\"./aiflow/notification-service/notify_new_member:\x01*\x12\xa4\x01\n\x0eregisterClient\x12+.notification_service.RegisterClientRequest\x1a,.notification_service.RegisterClientResponse\"7\x82\xd3\xe4\x93\x02\x31\",/aiflow/notification-service/register_client:\x01*\x12\x92\x01\n\x0c\x64\x65leteClient\x12%.notification_service.ClientIdRequest\x1a$.notification_service.CommonResponse\"5\x82\xd3\xe4\x93\x02/\"*/aiflow/notification-service/delete_client:\x01*\x12\x9f\x01\n\x0eisClientExists\x12%.notification_service.ClientIdRequest\x1a,.notification_service.isClientExistsResponse\"8\x82\xd3\xe4\x93\x02\x32\"-/aiflow/notification-service/is_client_exists:\x01*\x12\xb2\x01\n\x16timestampToEventOffset\x12).notification_service.TimeToOffsetRequest\x1a*.notification_service.TimeToOffsetResponse\"A\x82\xd3\xe4\x93\x02;\"6/aiflow/notification-service/timestamp_to_event_offset:\x01*B<\n\x1dorg.aiflow.notification.protoZ\x15/notification_service\x88\x01\x01\x90\x01\x01\x62\x06proto3'
  ,
  dependencies=[google_dot_api_dot_annotations__pb2.DESCRIPTOR,])

_RETURNSTATUS = _descriptor.EnumDescriptor(
  name='ReturnStatus',
  full_name='notification_service.ReturnStatus',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SUCCESS', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2344,
  serialized_end=2382,
)
_sym_db.RegisterEnumDescriptor(_RETURNSTATUS)

ReturnStatus = enum_type_wrapper.EnumTypeWrapper(_RETURNSTATUS)
SUCCESS = 0
ERROR = 1



_EVENTPROTO = _descriptor.Descriptor(
  name='EventProto',
  full_name='notification_service.EventProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='notification_service.EventProto.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='notification_service.EventProto.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='context', full_name='notification_service.EventProto.context', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='namespace', full_name='notification_service.EventProto.namespace', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='offset', full_name='notification_service.EventProto.offset', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='notification_service.EventProto.create_time', index=5,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sender', full_name='notification_service.EventProto.sender', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=83,
  serialized_end=212,
)


_SENDEREVENTCOUNTPROTO = _descriptor.Descriptor(
  name='SenderEventCountProto',
  full_name='notification_service.SenderEventCountProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='sender', full_name='notification_service.SenderEventCountProto.sender', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='event_count', full_name='notification_service.SenderEventCountProto.event_count', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=214,
  serialized_end=274,
)


_SENDEVENTREQUEST = _descriptor.Descriptor(
  name='SendEventRequest',
  full_name='notification_service.SendEventRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='event', full_name='notification_service.SendEventRequest.event', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='uuid', full_name='notification_service.SendEventRequest.uuid', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='enable_idempotence', full_name='notification_service.SendEventRequest.enable_idempotence', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=276,
  serialized_end=385,
)


_CLIENTMETA = _descriptor.Descriptor(
  name='ClientMeta',
  full_name='notification_service.ClientMeta',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='namespace', full_name='notification_service.ClientMeta.namespace', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sender', full_name='notification_service.ClientMeta.sender', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=387,
  serialized_end=434,
)


_SENDEVENTSRESPONSE = _descriptor.Descriptor(
  name='SendEventsResponse',
  full_name='notification_service.SendEventsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='event', full_name='notification_service.SendEventsResponse.event', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='return_code', full_name='notification_service.SendEventsResponse.return_code', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='return_msg', full_name='notification_service.SendEventsResponse.return_msg', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=437,
  serialized_end=583,
)


_LISTEVENTSREQUEST = _descriptor.Descriptor(
  name='ListEventsRequest',
  full_name='notification_service.ListEventsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='notification_service.ListEventsRequest.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='namespace', full_name='notification_service.ListEventsRequest.namespace', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sender', full_name='notification_service.ListEventsRequest.sender', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_offset', full_name='notification_service.ListEventsRequest.start_offset', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_offset', full_name='notification_service.ListEventsRequest.end_offset', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timeout_seconds', full_name='notification_service.ListEventsRequest.timeout_seconds', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=586,
  serialized_end=720,
)


_COUNTEVENTSREQUEST = _descriptor.Descriptor(
  name='CountEventsRequest',
  full_name='notification_service.CountEventsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='notification_service.CountEventsRequest.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='namespace', full_name='notification_service.CountEventsRequest.namespace', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sender', full_name='notification_service.CountEventsRequest.sender', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_offset', full_name='notification_service.CountEventsRequest.start_offset', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_offset', full_name='notification_service.CountEventsRequest.end_offset', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=722,
  serialized_end=832,
)


_LISTALLEVENTSREQUEST = _descriptor.Descriptor(
  name='ListAllEventsRequest',
  full_name='notification_service.ListAllEventsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timeout_seconds', full_name='notification_service.ListAllEventsRequest.timeout_seconds', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='notification_service.ListAllEventsRequest.start_time', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_offset', full_name='notification_service.ListAllEventsRequest.start_offset', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_offset', full_name='notification_service.ListAllEventsRequest.end_offset', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=834,
  serialized_end=943,
)


_LISTEVENTSRESPONSE = _descriptor.Descriptor(
  name='ListEventsResponse',
  full_name='notification_service.ListEventsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='return_code', full_name='notification_service.ListEventsResponse.return_code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='return_msg', full_name='notification_service.ListEventsResponse.return_msg', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='events', full_name='notification_service.ListEventsResponse.events', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=946,
  serialized_end=1093,
)


_COUNTEVENTSRESPONSE = _descriptor.Descriptor(
  name='CountEventsResponse',
  full_name='notification_service.CountEventsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='return_code', full_name='notification_service.CountEventsResponse.return_code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='return_msg', full_name='notification_service.CountEventsResponse.return_msg', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='event_count', full_name='notification_service.CountEventsResponse.event_count', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sender_event_counts', full_name='notification_service.CountEventsResponse.sender_event_counts', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1096,
  serialized_end=1289,
)


_MEMBERPROTO = _descriptor.Descriptor(
  name='MemberProto',
  full_name='notification_service.MemberProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='notification_service.MemberProto.version', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='server_uri', full_name='notification_service.MemberProto.server_uri', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='notification_service.MemberProto.update_time', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1291,
  serialized_end=1362,
)


_NOTIFY = _descriptor.Descriptor(
  name='Notify',
  full_name='notification_service.Notify',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='notification_service.Notify.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='namespace', full_name='notification_service.Notify.namespace', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1364,
  serialized_end=1404,
)


_NOTIFYREQUEST = _descriptor.Descriptor(
  name='NotifyRequest',
  full_name='notification_service.NotifyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='notifies', full_name='notification_service.NotifyRequest.notifies', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1406,
  serialized_end=1469,
)


_LISTMEMBERSREQUEST = _descriptor.Descriptor(
  name='ListMembersRequest',
  full_name='notification_service.ListMembersRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timeout_seconds', full_name='notification_service.ListMembersRequest.timeout_seconds', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1471,
  serialized_end=1516,
)


_LISTMEMBERSRESPONSE = _descriptor.Descriptor(
  name='ListMembersResponse',
  full_name='notification_service.ListMembersResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='return_code', full_name='notification_service.ListMembersResponse.return_code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='return_msg', full_name='notification_service.ListMembersResponse.return_msg', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='members', full_name='notification_service.ListMembersResponse.members', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1519,
  serialized_end=1669,
)


_NOTIFYNEWMEMBERREQUEST = _descriptor.Descriptor(
  name='NotifyNewMemberRequest',
  full_name='notification_service.NotifyNewMemberRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='member', full_name='notification_service.NotifyNewMemberRequest.member', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1671,
  serialized_end=1746,
)


_REGISTERCLIENTREQUEST = _descriptor.Descriptor(
  name='RegisterClientRequest',
  full_name='notification_service.RegisterClientRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='client_meta', full_name='notification_service.RegisterClientRequest.client_meta', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1748,
  serialized_end=1826,
)


_REGISTERCLIENTRESPONSE = _descriptor.Descriptor(
  name='RegisterClientResponse',
  full_name='notification_service.RegisterClientResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='return_code', full_name='notification_service.RegisterClientResponse.return_code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='return_msg', full_name='notification_service.RegisterClientResponse.return_msg', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_id', full_name='notification_service.RegisterClientResponse.client_id', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1828,
  serialized_end=1948,
)


_CLIENTIDREQUEST = _descriptor.Descriptor(
  name='ClientIdRequest',
  full_name='notification_service.ClientIdRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='client_id', full_name='notification_service.ClientIdRequest.client_id', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1950,
  serialized_end=1986,
)


_COMMONRESPONSE = _descriptor.Descriptor(
  name='CommonResponse',
  full_name='notification_service.CommonResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='return_code', full_name='notification_service.CommonResponse.return_code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='return_msg', full_name='notification_service.CommonResponse.return_msg', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1988,
  serialized_end=2081,
)


_ISCLIENTEXISTSRESPONSE = _descriptor.Descriptor(
  name='isClientExistsResponse',
  full_name='notification_service.isClientExistsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='return_code', full_name='notification_service.isClientExistsResponse.return_code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='return_msg', full_name='notification_service.isClientExistsResponse.return_msg', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_exists', full_name='notification_service.isClientExistsResponse.is_exists', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2083,
  serialized_end=2203,
)


_TIMETOOFFSETREQUEST = _descriptor.Descriptor(
  name='TimeToOffsetRequest',
  full_name='notification_service.TimeToOffsetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='notification_service.TimeToOffsetRequest.timestamp', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2205,
  serialized_end=2245,
)


_TIMETOOFFSETRESPONSE = _descriptor.Descriptor(
  name='TimeToOffsetResponse',
  full_name='notification_service.TimeToOffsetResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='return_code', full_name='notification_service.TimeToOffsetResponse.return_code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='offset', full_name='notification_service.TimeToOffsetResponse.offset', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2247,
  serialized_end=2342,
)

_SENDEVENTREQUEST.fields_by_name['event'].message_type = _EVENTPROTO
_SENDEVENTSRESPONSE.fields_by_name['event'].message_type = _EVENTPROTO
_SENDEVENTSRESPONSE.fields_by_name['return_code'].enum_type = _RETURNSTATUS
_LISTEVENTSRESPONSE.fields_by_name['return_code'].enum_type = _RETURNSTATUS
_LISTEVENTSRESPONSE.fields_by_name['events'].message_type = _EVENTPROTO
_COUNTEVENTSRESPONSE.fields_by_name['return_code'].enum_type = _RETURNSTATUS
_COUNTEVENTSRESPONSE.fields_by_name['sender_event_counts'].message_type = _SENDEREVENTCOUNTPROTO
_NOTIFYREQUEST.fields_by_name['notifies'].message_type = _NOTIFY
_LISTMEMBERSRESPONSE.fields_by_name['return_code'].enum_type = _RETURNSTATUS
_LISTMEMBERSRESPONSE.fields_by_name['members'].message_type = _MEMBERPROTO
_NOTIFYNEWMEMBERREQUEST.fields_by_name['member'].message_type = _MEMBERPROTO
_REGISTERCLIENTREQUEST.fields_by_name['client_meta'].message_type = _CLIENTMETA
_REGISTERCLIENTRESPONSE.fields_by_name['return_code'].enum_type = _RETURNSTATUS
_COMMONRESPONSE.fields_by_name['return_code'].enum_type = _RETURNSTATUS
_ISCLIENTEXISTSRESPONSE.fields_by_name['return_code'].enum_type = _RETURNSTATUS
_TIMETOOFFSETRESPONSE.fields_by_name['return_code'].enum_type = _RETURNSTATUS
DESCRIPTOR.message_types_by_name['EventProto'] = _EVENTPROTO
DESCRIPTOR.message_types_by_name['SenderEventCountProto'] = _SENDEREVENTCOUNTPROTO
DESCRIPTOR.message_types_by_name['SendEventRequest'] = _SENDEVENTREQUEST
DESCRIPTOR.message_types_by_name['ClientMeta'] = _CLIENTMETA
DESCRIPTOR.message_types_by_name['SendEventsResponse'] = _SENDEVENTSRESPONSE
DESCRIPTOR.message_types_by_name['ListEventsRequest'] = _LISTEVENTSREQUEST
DESCRIPTOR.message_types_by_name['CountEventsRequest'] = _COUNTEVENTSREQUEST
DESCRIPTOR.message_types_by_name['ListAllEventsRequest'] = _LISTALLEVENTSREQUEST
DESCRIPTOR.message_types_by_name['ListEventsResponse'] = _LISTEVENTSRESPONSE
DESCRIPTOR.message_types_by_name['CountEventsResponse'] = _COUNTEVENTSRESPONSE
DESCRIPTOR.message_types_by_name['MemberProto'] = _MEMBERPROTO
DESCRIPTOR.message_types_by_name['Notify'] = _NOTIFY
DESCRIPTOR.message_types_by_name['NotifyRequest'] = _NOTIFYREQUEST
DESCRIPTOR.message_types_by_name['ListMembersRequest'] = _LISTMEMBERSREQUEST
DESCRIPTOR.message_types_by_name['ListMembersResponse'] = _LISTMEMBERSRESPONSE
DESCRIPTOR.message_types_by_name['NotifyNewMemberRequest'] = _NOTIFYNEWMEMBERREQUEST
DESCRIPTOR.message_types_by_name['RegisterClientRequest'] = _REGISTERCLIENTREQUEST
DESCRIPTOR.message_types_by_name['RegisterClientResponse'] = _REGISTERCLIENTRESPONSE
DESCRIPTOR.message_types_by_name['ClientIdRequest'] = _CLIENTIDREQUEST
DESCRIPTOR.message_types_by_name['CommonResponse'] = _COMMONRESPONSE
DESCRIPTOR.message_types_by_name['isClientExistsResponse'] = _ISCLIENTEXISTSRESPONSE
DESCRIPTOR.message_types_by_name['TimeToOffsetRequest'] = _TIMETOOFFSETREQUEST
DESCRIPTOR.message_types_by_name['TimeToOffsetResponse'] = _TIMETOOFFSETRESPONSE
DESCRIPTOR.enum_types_by_name['ReturnStatus'] = _RETURNSTATUS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EventProto = _reflection.GeneratedProtocolMessageType('EventProto', (_message.Message,), {
  'DESCRIPTOR' : _EVENTPROTO,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.EventProto)
  })
_sym_db.RegisterMessage(EventProto)

SenderEventCountProto = _reflection.GeneratedProtocolMessageType('SenderEventCountProto', (_message.Message,), {
  'DESCRIPTOR' : _SENDEREVENTCOUNTPROTO,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.SenderEventCountProto)
  })
_sym_db.RegisterMessage(SenderEventCountProto)

SendEventRequest = _reflection.GeneratedProtocolMessageType('SendEventRequest', (_message.Message,), {
  'DESCRIPTOR' : _SENDEVENTREQUEST,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.SendEventRequest)
  })
_sym_db.RegisterMessage(SendEventRequest)

ClientMeta = _reflection.GeneratedProtocolMessageType('ClientMeta', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTMETA,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.ClientMeta)
  })
_sym_db.RegisterMessage(ClientMeta)

SendEventsResponse = _reflection.GeneratedProtocolMessageType('SendEventsResponse', (_message.Message,), {
  'DESCRIPTOR' : _SENDEVENTSRESPONSE,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.SendEventsResponse)
  })
_sym_db.RegisterMessage(SendEventsResponse)

ListEventsRequest = _reflection.GeneratedProtocolMessageType('ListEventsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTEVENTSREQUEST,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.ListEventsRequest)
  })
_sym_db.RegisterMessage(ListEventsRequest)

CountEventsRequest = _reflection.GeneratedProtocolMessageType('CountEventsRequest', (_message.Message,), {
  'DESCRIPTOR' : _COUNTEVENTSREQUEST,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.CountEventsRequest)
  })
_sym_db.RegisterMessage(CountEventsRequest)

ListAllEventsRequest = _reflection.GeneratedProtocolMessageType('ListAllEventsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTALLEVENTSREQUEST,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.ListAllEventsRequest)
  })
_sym_db.RegisterMessage(ListAllEventsRequest)

ListEventsResponse = _reflection.GeneratedProtocolMessageType('ListEventsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTEVENTSRESPONSE,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.ListEventsResponse)
  })
_sym_db.RegisterMessage(ListEventsResponse)

CountEventsResponse = _reflection.GeneratedProtocolMessageType('CountEventsResponse', (_message.Message,), {
  'DESCRIPTOR' : _COUNTEVENTSRESPONSE,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.CountEventsResponse)
  })
_sym_db.RegisterMessage(CountEventsResponse)

MemberProto = _reflection.GeneratedProtocolMessageType('MemberProto', (_message.Message,), {
  'DESCRIPTOR' : _MEMBERPROTO,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.MemberProto)
  })
_sym_db.RegisterMessage(MemberProto)

Notify = _reflection.GeneratedProtocolMessageType('Notify', (_message.Message,), {
  'DESCRIPTOR' : _NOTIFY,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.Notify)
  })
_sym_db.RegisterMessage(Notify)

NotifyRequest = _reflection.GeneratedProtocolMessageType('NotifyRequest', (_message.Message,), {
  'DESCRIPTOR' : _NOTIFYREQUEST,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.NotifyRequest)
  })
_sym_db.RegisterMessage(NotifyRequest)

ListMembersRequest = _reflection.GeneratedProtocolMessageType('ListMembersRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTMEMBERSREQUEST,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.ListMembersRequest)
  })
_sym_db.RegisterMessage(ListMembersRequest)

ListMembersResponse = _reflection.GeneratedProtocolMessageType('ListMembersResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTMEMBERSRESPONSE,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.ListMembersResponse)
  })
_sym_db.RegisterMessage(ListMembersResponse)

NotifyNewMemberRequest = _reflection.GeneratedProtocolMessageType('NotifyNewMemberRequest', (_message.Message,), {
  'DESCRIPTOR' : _NOTIFYNEWMEMBERREQUEST,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.NotifyNewMemberRequest)
  })
_sym_db.RegisterMessage(NotifyNewMemberRequest)

RegisterClientRequest = _reflection.GeneratedProtocolMessageType('RegisterClientRequest', (_message.Message,), {
  'DESCRIPTOR' : _REGISTERCLIENTREQUEST,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.RegisterClientRequest)
  })
_sym_db.RegisterMessage(RegisterClientRequest)

RegisterClientResponse = _reflection.GeneratedProtocolMessageType('RegisterClientResponse', (_message.Message,), {
  'DESCRIPTOR' : _REGISTERCLIENTRESPONSE,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.RegisterClientResponse)
  })
_sym_db.RegisterMessage(RegisterClientResponse)

ClientIdRequest = _reflection.GeneratedProtocolMessageType('ClientIdRequest', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTIDREQUEST,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.ClientIdRequest)
  })
_sym_db.RegisterMessage(ClientIdRequest)

CommonResponse = _reflection.GeneratedProtocolMessageType('CommonResponse', (_message.Message,), {
  'DESCRIPTOR' : _COMMONRESPONSE,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.CommonResponse)
  })
_sym_db.RegisterMessage(CommonResponse)

isClientExistsResponse = _reflection.GeneratedProtocolMessageType('isClientExistsResponse', (_message.Message,), {
  'DESCRIPTOR' : _ISCLIENTEXISTSRESPONSE,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.isClientExistsResponse)
  })
_sym_db.RegisterMessage(isClientExistsResponse)

TimeToOffsetRequest = _reflection.GeneratedProtocolMessageType('TimeToOffsetRequest', (_message.Message,), {
  'DESCRIPTOR' : _TIMETOOFFSETREQUEST,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.TimeToOffsetRequest)
  })
_sym_db.RegisterMessage(TimeToOffsetRequest)

TimeToOffsetResponse = _reflection.GeneratedProtocolMessageType('TimeToOffsetResponse', (_message.Message,), {
  'DESCRIPTOR' : _TIMETOOFFSETRESPONSE,
  '__module__' : 'notification_service_pb2'
  # @@protoc_insertion_point(class_scope:notification_service.TimeToOffsetResponse)
  })
_sym_db.RegisterMessage(TimeToOffsetResponse)


DESCRIPTOR._options = None

_NOTIFICATIONSERVICE = _descriptor.ServiceDescriptor(
  name='NotificationService',
  full_name='notification_service.NotificationService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=2385,
  serialized_end=4120,
  methods=[
  _descriptor.MethodDescriptor(
    name='sendEvent',
    full_name='notification_service.NotificationService.sendEvent',
    index=0,
    containing_service=None,
    input_type=_SENDEVENTREQUEST,
    output_type=_SENDEVENTSRESPONSE,
    serialized_options=b'\202\323\344\223\002,\"\'/aiflow/notification-service/send_event:\001*',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='listEvents',
    full_name='notification_service.NotificationService.listEvents',
    index=1,
    containing_service=None,
    input_type=_LISTEVENTSREQUEST,
    output_type=_LISTEVENTSRESPONSE,
    serialized_options=b'\202\323\344\223\002*\022(/aiflow/notification-service/list_events',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='countEvents',
    full_name='notification_service.NotificationService.countEvents',
    index=2,
    containing_service=None,
    input_type=_COUNTEVENTSREQUEST,
    output_type=_COUNTEVENTSRESPONSE,
    serialized_options=b'\202\323\344\223\002+\022)/aiflow/notification-service/count_events',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='listAllEvents',
    full_name='notification_service.NotificationService.listAllEvents',
    index=3,
    containing_service=None,
    input_type=_LISTALLEVENTSREQUEST,
    output_type=_LISTEVENTSRESPONSE,
    serialized_options=b'\202\323\344\223\002.\022,/aiflow/notification-service/list_all_events',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='notify',
    full_name='notification_service.NotificationService.notify',
    index=4,
    containing_service=None,
    input_type=_NOTIFYREQUEST,
    output_type=_COMMONRESPONSE,
    serialized_options=b'\202\323\344\223\002(\"#/aiflow/notification-service/notify:\001*',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='listMembers',
    full_name='notification_service.NotificationService.listMembers',
    index=5,
    containing_service=None,
    input_type=_LISTMEMBERSREQUEST,
    output_type=_LISTMEMBERSRESPONSE,
    serialized_options=b'\202\323\344\223\002+\022)/aiflow/notification-service/list_members',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='notifyNewMember',
    full_name='notification_service.NotificationService.notifyNewMember',
    index=6,
    containing_service=None,
    input_type=_NOTIFYNEWMEMBERREQUEST,
    output_type=_COMMONRESPONSE,
    serialized_options=b'\202\323\344\223\0023\"./aiflow/notification-service/notify_new_member:\001*',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='registerClient',
    full_name='notification_service.NotificationService.registerClient',
    index=7,
    containing_service=None,
    input_type=_REGISTERCLIENTREQUEST,
    output_type=_REGISTERCLIENTRESPONSE,
    serialized_options=b'\202\323\344\223\0021\",/aiflow/notification-service/register_client:\001*',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='deleteClient',
    full_name='notification_service.NotificationService.deleteClient',
    index=8,
    containing_service=None,
    input_type=_CLIENTIDREQUEST,
    output_type=_COMMONRESPONSE,
    serialized_options=b'\202\323\344\223\002/\"*/aiflow/notification-service/delete_client:\001*',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='isClientExists',
    full_name='notification_service.NotificationService.isClientExists',
    index=9,
    containing_service=None,
    input_type=_CLIENTIDREQUEST,
    output_type=_ISCLIENTEXISTSRESPONSE,
    serialized_options=b'\202\323\344\223\0022\"-/aiflow/notification-service/is_client_exists:\001*',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='timestampToEventOffset',
    full_name='notification_service.NotificationService.timestampToEventOffset',
    index=10,
    containing_service=None,
    input_type=_TIMETOOFFSETREQUEST,
    output_type=_TIMETOOFFSETRESPONSE,
    serialized_options=b'\202\323\344\223\002;\"6/aiflow/notification-service/timestamp_to_event_offset:\001*',
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_NOTIFICATIONSERVICE)

DESCRIPTOR.services_by_name['NotificationService'] = _NOTIFICATIONSERVICE

NotificationService = service_reflection.GeneratedServiceType('NotificationService', (_service.Service,), dict(
  DESCRIPTOR = _NOTIFICATIONSERVICE,
  __module__ = 'notification_service_pb2'
  ))

NotificationService_Stub = service_reflection.GeneratedServiceStubType('NotificationService_Stub', (NotificationService,), dict(
  DESCRIPTOR = _NOTIFICATIONSERVICE,
  __module__ = 'notification_service_pb2'
  ))


# @@protoc_insertion_point(module_scope)



#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import logging
import os
import signal
import socket
import time
from importlib import import_module
from itertools import tee, filterfalse

import dateutil.parser
from typing import Callable, Iterable

from notification_service.model.event import Event
from notification_service.model.member import Member
from notification_service.model.sender_event_count import SenderEventCount
from notification_service.rpc.protobuf import notification_service_pb2

if not hasattr(time, 'time_ns'):
    time.time_ns = lambda: int(time.time() * 1e9)

logger = logging.getLogger(__name__)


def event_to_proto(event: Event):
    result_event_proto = notification_service_pb2.EventProto(key=event.key,
                                                             offset=event.offset,
                                                             value=event.value,
                                                             create_time=event.create_time,
                                                             namespace=event.namespace,
                                                             context=event.context,
                                                             sender=event.sender)
    return result_event_proto


def count_to_proto(count: SenderEventCount):
    sender_event_count_proto = notification_service_pb2.SenderEventCountProto(sender=count.sender,
                                                                              event_count=count.event_count)
    return sender_event_count_proto


def event_list_to_proto(event_list):
    event_proto_list = []
    for event_model in event_list:
        event_proto = event_to_proto(event_model)
        event_proto_list.append(event_proto)
    return event_proto_list


def count_list_to_proto(count_list):
    event_count = 0
    count_proto_list = []
    for count in count_list:
        event_count += count.event_count
        count_proto_list.append(count_to_proto(count))
    return event_count, count_proto_list


def event_proto_to_event(event_proto):
    event = Event(key=event_proto.key, value=event_proto.value)
    event.namespace = event_proto.namespace
    event.sender = event_proto.sender
    event.context = event_proto.context
    event.offset = event_proto.offset
    event.create_time = event_proto.create_time
    return event


def event_count_proto_to_event_count(event_count_proto):
    return SenderEventCount(sender=event_count_proto.sender,
                            event_count=event_count_proto.event_count)


def event_model_to_event(event_model):
    event = Event(key=event_model.key, value=event_model.value)
    event.namespace = event_model.namespace
    event.sender = event_model.sender
    event.offset = event_model.offset
    event.create_time = event_model.create_time
    event.context = event_model.context
    return event


def count_result_to_sender_event_count(count_result):
    return SenderEventCount(
        sender=count_result.sender,
        event_count=count_result.event_count
    )


def member_to_proto(member: Member):
    return notification_service_pb2.MemberProto(
        version=member.version, server_uri=member.server_uri, update_time=member.update_time)


def proto_to_member(member_proto):
    return Member(member_proto.version, member_proto.server_uri, member_proto.update_time)


def sleep_and_detecting_running(interval_ms, is_running_callable, min_interval_ms=500):
    start_time = time.monotonic() * 1000
    while is_running_callable() and time.monotonic() * 1000 < start_time + interval_ms:
        remaining = time.monotonic() * 1000 - start_time
        if remaining > min_interval_ms:
            time.sleep(min_interval_ms / 1000)
        else:
            time.sleep(remaining / 1000)


def get_ip_addr() -> str:
    """
    Get ip address of localhost by UDP socket.
    :return: ip address of localhost
    """
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    except OSError:
        ip = '127.0.0.1'
    finally:
        s.close()
        return ip


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:
        raise ImportError(f"{dotted_path} doesn't look like a module path")

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError:
        raise ImportError(f'Module "{module_path}" does not define a "{class_name}" attribute/class')


def partition(pred: Callable, iterable: Iterable):
    """Use a predicate to partition entries into false entries and true entries"""
    iter_1, iter_2 = tee(iterable)
    return filterfalse(pred, iter_1), filter(pred, iter_2)


def parse_date(dt: str):
    """
    Parse a datetime in ISO 8601 to milliseconds since epoch.

    :param dt: datetime string
    """
    parsed_time = dateutil.parser.parse(dt)
    return int(parsed_time.timestamp() * 1000)


def check_pid_exist(_pid):
    try:
        os.kill(_pid, 0)
    except OSError:
        return False
    else:
        return True


def stop_process(pid, process_name, timeout_sec: int = 60):
    """
    Try hard to kill the process with the given pid. It sends
    :param pid: The process pid to stop
    :param process_name: The process name to log
    :param timeout_sec: timeout before sending SIGKILL to kill the process
    """

    try:
        start_time = time.monotonic()
        while check_pid_exist(pid):
            if time.monotonic() - start_time > timeout_sec:
                raise RuntimeError(
                    "{} pid: {} does not exit after {} seconds.".format(process_name, pid, timeout_sec))
            os.kill(pid, signal.SIGTERM)
            time.sleep(1)
    except Exception:
        logger.warning("Failed to stop {} pid: {} with SIGTERM. Try to send SIGKILL".format(process_name, pid))
        try:
            os.kill(pid, signal.SIGKILL)
        except Exception as e:
            raise RuntimeError("Failed to kill {} pid: {} with SIGKILL.".format(process_name, pid)) from e

    logger.info("{} pid: {} stopped".format(process_name, pid))



#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import contextlib
import re
import time
import logging
import os
import urllib.parse
import sqlalchemy
from enum import Enum
from functools import wraps

from sqlalchemy import create_engine, Column, String, BigInteger, Text, Integer, Boolean
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session, sessionmaker


if not hasattr(time, 'time_ns'):
    time.time_ns = lambda: int(time.time() * 1e9)

# use sqlite by default for testing
SQL_ALCHEMY_DB_FILE = 'notification_service.db'
SQL_ALCHEMY_CONN = "sqlite:///" + SQL_ALCHEMY_DB_FILE
engine = None
Session = None

_logger = logging.getLogger(__name__)


"""
List of SQLAlchemy database engines for notification backend storage.
"""

MYSQL = 'mysql'
SQLITE = 'sqlite'
POSTGRES = 'postgresql'
MSSQL = 'mssql'
MONGODB = 'mongodb'

DATABASE_ENGINES = [
    MYSQL,
    SQLITE,
    POSTGRES,
    MSSQL,
    MONGODB,
]


class DBType(str, Enum):
    SQLITE = "sql_lite"
    MYSQL = "mysql"
    MONGODB = "mongodb"

    @staticmethod
    def value_of(db_type):
        if db_type in ('SQL_LITE', 'sql_lite', 'sqlite'):
            return DBType.SQLITE
        elif db_type in ('MYSQL', 'mysql'):
            return DBType.MYSQL
        elif db_type in ('MONGODB', 'mongodb'):
            return DBType.MONGODB
        else:
            raise NotImplementedError


def _get_alembic_config(url):
    from alembic.config import Config

    current_dir = os.path.dirname(os.path.abspath(__file__))
    package_dir = os.path.normpath(os.path.join(current_dir, '..'))
    directory = os.path.join(package_dir, 'migrations')
    config = Config(os.path.join(package_dir, 'alembic.ini'))
    config.set_main_option('script_location', directory.replace('%', '%%'))
    config.set_main_option('sqlalchemy.url', url.replace('%', '%%'))
    return config


def upgrade(url, version=None):
    """Upgrade the database."""
    # alembic adds significant import time, so we import it lazily
    from alembic import command

    _logger.info('Upgrade the database, db uri: {}'.format(url))
    config = _get_alembic_config(url)
    if version is None:
        version = 'heads'
    command.upgrade(config, version)


def clear_db(url, metadata):
    _logger.info('Clear the database, db uri: {}'.format(url))
    engine = sqlalchemy.create_engine(url)
    connection = engine.connect()
    metadata.drop_all(connection)
    # alembic adds significant import time, so we import it lazily
    from alembic.migration import MigrationContext  # noqa

    migration_ctx = MigrationContext.configure(connection)
    version = migration_ctx._version  # noqa pylint: disable=protected-access
    if version.exists(connection):
        version.drop(connection)


def reset_db(url, metadata):
    _logger.info('Reset the database, db uri: {}'.format(url))
    clear_db(url, metadata)
    upgrade(url)


def downgrade(url, version):
    """Upgrade the database."""
    # alembic adds significant import time, so we import it lazily
    from alembic import command

    _logger.info('Downgrade the database, db uri: {}'.format(url))
    config = _get_alembic_config(url)
    command.downgrade(config, version)


def tables_exists(url):
    """Returns whether tables are created"""
    engine_ = create_engine(url)
    if len(engine_.table_names()) > 0:
        return True
    else:
        return False


def extract_db_engine_from_uri(db_uri):
    """
    Parse specified database URI to extract database type. Confirm extracted database engine is
    supported. If database driver is specified, confirm driver passes a plausible regex.
    """
    scheme = urllib.parse.urlparse(db_uri).scheme
    scheme_plus_count = scheme.count('+')

    """validates scheme parsed from DB URI is supported"""
    if scheme_plus_count == 0:
        db_engine = scheme
    elif scheme_plus_count == 1:
        db_engine, _ = scheme.split('+')
    else:
        error_msg = "Invalid database URI: '%s'." % db_uri
        raise Exception(error_msg)

    """validates db_engine parsed from DB URI is supported"""
    if db_engine not in DATABASE_ENGINES:
        error_msg = "Invalid database engine: '%s'." % db_engine
        raise Exception(error_msg)

    return db_engine


def parse_mongo_uri(db_uri):
    """
    Parse MongoDB URI-style string to split up and return credentials

    Args:
        db_uri (string): MongoDB URI-style string
    Return:

    """
    regex_str = r'^(?P<schema>(mongodb:(?:\/{2})?))((?P<user>\w+?):(?P<pwd>(\w+?))@|:@?)(?P<host>(\S+?)):(?P<port>(\d+))(\/(?P<db>(\S+?)))$'
    pattern = re.compile(regex_str)
    m = pattern.match(db_uri)
    if m is None:
        raise Exception('The URI of MongoDB is invalid')
    return m.group('user'), m.group('pwd'), m.group('host'), m.group('port'), m.group('db')


def prepare_db(user_engine=None, user_session=None, print_sql=False):
    global engine
    global Session
    if user_engine is not None and user_session is not None:
        engine = user_engine
        Session = user_session
    if engine is None or Session is None:
        engine_args = {'encoding': "utf-8"}
        if print_sql:
            engine_args['echo'] = True
        engine = create_engine(SQL_ALCHEMY_CONN, **engine_args)
        Session = scoped_session(
            sessionmaker(autocommit=False,
                         autoflush=False,
                         bind=engine,
                         expire_on_commit=False))


def clear_engine_and_session():
    global engine
    global Session
    engine = None
    Session = None


@contextlib.contextmanager
def create_session():
    prepare_db()
    session = Session()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


def provide_session(func):
    """
    Function decorator that provides a session if it isn't provided.
    If you want to reuse a session or run the function as part of a
    database transaction, you pass it to the function, if not this wrapper
    will create one and close it for you.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        arg_session = 'session'

        func_params = func.__code__.co_varnames
        session_in_args = arg_session in func_params and \
                          func_params.index(arg_session) < len(args)
        session_in_kwargs = arg_session in kwargs

        if session_in_kwargs or session_in_args:
            return func(*args, **kwargs)
        else:
            with create_session() as session:
                kwargs[arg_session] = session
                return func(*args, **kwargs)

    return wrapper


def create_all_tables(db_conn=None):
    global SQL_ALCHEMY_CONN
    if db_conn is not None:
        SQL_ALCHEMY_CONN = db_conn
    if not tables_exists(SQL_ALCHEMY_CONN):
        upgrade(url=SQL_ALCHEMY_CONN)
    prepare_db()


# def drop_all_tables(db_conn=None):
#     global SQL_ALCHEMY_CONN
#     if db_conn is not None:
#         SQL_ALCHEMY_CONN = db_conn
#     clear_db(url=SQL_ALCHEMY_CONN)



# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""DB command"""
import logging

from notification_service.settings import get_configuration
from notification_service.storage.alchemy.base import Base
from notification_service.util import db

_logger = logging.getLogger(__name__)


def init(args):
    """Init the metadata database"""
    config = get_configuration()
    _logger.info('Initialize the database, db uri: {}'.format(config.db_uri))
    db.upgrade(url=config.db_uri)


def reset(args):
    """Reset the metadata database"""
    config = get_configuration()
    _logger.info('Reset the database, db uri: {}'.format(config.db_uri))
    if args.yes or input("This will drop existing tables if they exist. Proceed? (y/n)").upper() == "Y":
        db.reset_db(url=config.db_uri, metadata=Base.metadata)
    else:
        _logger.info('Cancel reset the database, db uri: {}'.format(config.db_uri))


def upgrade(args):
    """Upgrade the metadata database"""
    config = get_configuration()
    _logger.info('Upgrade the database, db uri: {}'.format(config.db_uri))
    db.upgrade(url=config.db_uri, version=args.version)


def downgrade(args):
    """Downgrade the metadata database"""
    config = get_configuration()
    _logger.info('Downgrade the database, db uri: {}'.format(config.db_uri))
    db.downgrade(url=config.db_uri, version=args.version)
